@@ -40,35 +40,135 @@ window.GoAccess = window.GoAccess || {
 		this.opts = options;
 		var cw = Math.max(document.documentElement.clientWidth || 0, window.innerWidth || 0);
 
-		this.AppState  = {}; // current state app key-value store
-		this.AppTpls   = {}; // precompiled templates
-		this.AppCharts = {}; // holds all rendered charts
-		this.AppUIData = (this.opts || {}).uiData || {};    // holds panel definitions
-		this.AppData   = (this.opts || {}).panelData || {}; // hold raw data
-		this.AppWSConn = (this.opts || {}).wsConnection || {}; // WebSocket connection
-		this.i18n = (this.opts || {}).i18n || {}; // i18n report labels
-		this.AppPrefs  = {
+		this.AppState = {};
+		this.AppTpls = {};
+		this.AppCharts = {};
+		this.AppUIData = (this.opts || {}).uiData || {};
+		this.AppData = (this.opts || {}).panelData || {};
+		this.AppWSConn = (this.opts || {}).wsConnection || {};
+		this.i18n = (this.opts || {}).i18n || {};
+		this.AppPrefs = {
 			'autoHideTables': true,
 			'layout': cw > 2560 ? 'wide' : 'horizontal',
+			'panelOrder': [],
 			'perPage': 7,
 			'theme': (window.matchMedia && window.matchMedia('(prefers-color-scheme: dark)').matches) ? 'darkPurple' : 'bright',
 			'hiddenPanels': [],
 		};
 		this.AppPrefs = GoAccess.Util.merge(this.AppPrefs, this.opts.prefs);
+		this.currentJWT = null;
+		this.csrfToken = null;
 
-		// WebSocket reconnection
-		this.wsDelay    = this.currDelay = 1E3;
-		this.maxDelay   = 20E3;
-		this.retries    = 0;
+		// WebSocket reconnection settings
+		this.wsDelay = this.currDelay = 1E3;
+		this.maxDelay = 20E3;
+		this.retries = 0;
 		this.maxRetries = 20;
+		this.tokenRefreshLeadTime = 60;
 
+		this.handleLocalStorage();
+		this.isAppInitialized = false;
+
+		// Initialize message rotation
+		this.startMessageRotation();
+
+		// Handle WebSocket setup
+		this.handleWebSocketSetup();
+	},
+
+	handleLocalStorage: function () {
+		// Check if the browser supports localStorage
 		if (GoAccess.Util.hasLocalStorage()) {
-			var ls = JSON.parse(localStorage.getItem('AppPrefs'));
+			// Retrieve the AppPrefs object stored in localStorage
+			const ls = JSON.parse(localStorage.getItem('AppPrefs'));
+			// Merge stored preferences into the current application preferences
 			this.AppPrefs = GoAccess.Util.merge(this.AppPrefs, ls);
 		}
-		if (Object.keys(this.AppWSConn).length)
-			this.setWebSocket(this.AppWSConn);
+	},
 
+	startMessageRotation: function () {
+		// Define the messages that will be displayed during the loading process
+		const messages = [
+			'Fetching authentication token... Please wait.',
+			'Validating WebSocket tokens... Please wait.',
+			'Authenticating WebSocket connection... Please wait.',
+			'Verifying WebSocket credentials... Please wait.',
+			'Authorizing WebSocket session... Please wait.'
+		];
+		let currentMessageIndex = 0; // Tracks the index of the currently displayed message
+		// Set up an interval to rotate through the messages every 100ms
+		this.messageInterval = setInterval(() => {
+			if (currentMessageIndex < messages.length) {
+				// Update the loading status element with the current message
+				$('.app-loading-status > small').innerHTML = messages[currentMessageIndex];
+				currentMessageIndex++; // Move to the next message
+			}
+		}, 500);
+	},
+
+	handleWebSocketSetup: function () {
+		// Fetch and authenticate the JWT using the provided WebSocket auth URL (external JWT)
+		if (this.AppWSConn.ws_auth_url) {
+			this.fetchAndAuthenticateJWT();
+		}
+		// If a JWT exists or WebSocket configuration is provided
+		else if (window.goaccessJWT || Object.keys(this.AppWSConn).length) {
+			// Set up the WebSocket connection using the existing JWT
+			this.setWebSocket(this.AppWSConn, window.goaccessJWT, this.messageInterval);
+		}  else {
+			// Initialize the application without WebSocket authentication
+			this.initializeWithoutWebSocket();
+		}
+	},
+
+	fetchAndAuthenticateJWT: function () {
+		// Attempt to fetch a JWT from the WebSocket authentication URL
+		this.fetchJWT(this.AppWSConn.ws_auth_url)
+			.then(data => {
+				if (data.status === "success") {
+					// Extract the JWT, refresh token, and expiration time from the response
+					const jwt = data.access_token;
+					const refreshToken = data.refresh_token;
+					const expiresIn = data.expires_in;
+					// Set up the WebSocket connection using the fetched JWT
+					this.setWebSocket(this.AppWSConn, jwt, this.messageInterval);
+					// Schedule automatic token refresh before it expires
+					this.scheduleTokenRefresh(expiresIn, refreshToken);
+				} else {
+					// Handle failure response from the authentication server
+					this.handleAuthenticationFailure(data.message);
+				}
+			})
+			.catch(error => {
+				// Handle errors during the JWT fetch process
+				this.handleAuthenticationError(error);
+			});
+	},
+
+	initializeWithoutWebSocket: function () {
+		// Stop the message rotation interval
+		clearInterval(this.messageInterval);
+		// Update the UI to indicate that no authentication is provided
+		$('.app-loading-status > small').innerHTML = 'No authentication provided.';
+		// Proceed to initialize the app without WebSocket support
+		GoAccess.App.initialize();
+		this.isAppInitialized = true;
+	},
+
+	handleAuthenticationFailure: function (message) {
+		// Stop the message rotation interval
+		clearInterval(this.messageInterval);
+		// Update the UI to display the failure message
+		$('.app-loading-status > small').innerHTML = `Authentication failed: ${message}`;
+		// Hide the loading spinner
+		$('.loading-container > .spinner').style.display = 'none';
+	},
+
+	handleAuthenticationError: function (error) {
+		// Stop the message rotation interval
+		clearInterval(this.messageInterval);
+		// Update the UI to indicate an error occurred during the JWT fetch process
+		$('.app-loading-status > small').innerHTML = 'Error fetching authentication token.';
 	},
 
 	getPanelUI: function (panel) {
@@ -89,14 +189,86 @@ window.GoAccess = window.GoAccess || {
 		return panel ? this.AppData[panel] : this.AppData;
 	},
 
+	// Include cookies for session validation
+	fetchJWT: function (url) {
+		return fetch(url, {
+			method: 'GET',
+			credentials: 'include',
+			headers: { 'Accept': 'application/json' },
+			referrerPolicy: 'no-referrer-when-downgrade'
+		})
+		.then(response => response.json())
+		.then(data => {
+			if (data.status === 'success' && data.csrf_token) {
+				this.csrfToken = data.csrf_token;
+			}
+			return data;
+		});
+	},
+
+	refreshJWT: function (url, refreshToken) {
+		const headers = {
+			'Accept': 'application/json',
+			'Content-Type': 'application/json'
+		};
+		if (this.csrfToken) {
+			headers['X-CSRF-TOKEN'] = this.csrfToken;
+		}
+		return fetch(url, {
+			method: 'POST',
+			credentials: 'include',
+			headers: headers,
+			referrerPolicy: 'no-referrer-when-downgrade',
+			body: JSON.stringify({ refresh_token: refreshToken })
+		}).then(response => response.json());
+	},
+
+	// Schedule the next token refresh, triggering a refresh shortly before the token expires
+	scheduleTokenRefresh: function (expiresIn, refreshToken) {
+		// Refresh 1 minute before expiration
+		const refreshUrl = this.AppWSConn.ws_auth_refresh_url || this.AppWSConn.ws_auth_url;
+		// Set the timer to trigger one minute before the token expires
+		setTimeout(() => {
+			this.refreshJWT(refreshUrl, refreshToken)
+				.then(data => {
+					if (data.status === "success") {
+						const newJwt = data.access_token;
+						const newRefreshToken = data.refresh_token;
+						const newExpiresIn = data.expires_in;
+						// Update token without reconnecting
+						this.sendNewJWT(newJwt);
+						// Schedule the next refresh using the new expiration time
+						this.scheduleTokenRefresh(newExpiresIn, newRefreshToken);
+					} else {
+						// Update token without reconnecting
+						this.sendNewJWT(null);
+					}
+				})
+				.catch(error => {
+					console.error("Error refreshing JWT:", error);
+				});
+		}, (expiresIn - this.tokenRefreshLeadTime) * 1000);
+	},
+
+	// Sends the new JWT to the server over the already-open WebSocket connection
+	sendNewJWT: function (newJwt) {
+		if (this.socket && this.socket.readyState === WebSocket.OPEN) {
+			// Notify the server to update the JWT used for authentication
+			this.socket.send(JSON.stringify({ action: "validate_token", token: newJwt }));
+		}
+		// Also update the locally stored token
+		this.currentJWT = newJwt;
+	},
+
 	reconnect: function (wsConn) {
 		if (this.retries >= this.maxRetries)
 			return window.clearTimeout(this.wsTimer);
 
 		this.retries++;
+		// Exponential backoff
 		if (this.currDelay < this.maxDelay)
-			this.currDelay *= 2; // Exponential backoff
-		this.setWebSocket(wsConn);
+			this.currDelay *= 2;
+		this.setWebSocket(wsConn, this.currentJWT, null);
 	},
 
 	buildWSURI: function (wsConn) {
@@ -107,8 +279,27 @@ window.GoAccess = window.GoAccess || {
 		return new URL(url).protocol + '//' + new URL(url).hostname + ':' + wsConn.port + new URL(url).pathname;
 	},
 
-	setWebSocket: function (wsConn) {
+	setWebSocket: function (wsConn, jwt, messageInterval) {
 		var host = null, pingId = null, uri = null, defURI = null, str = null;
+		// Store the JWT used for this connection
+		this.currentJWT = jwt;
+
+		// If no external messageInterval is provided, set up local message rotation
+		if (jwt && !messageInterval) {
+			const messages = [
+				'Validating WebSocket tokens... Please wait.',
+				'Authenticating WebSocket connection... Please wait.',
+				'Verifying WebSocket credentials... Please wait.',
+				'Authorizing WebSocket session... Please wait.'
+			];
+			let currentMessageIndex = 0;
+			messageInterval = setInterval(() => {
+				if (currentMessageIndex < messages.length) {
+					$('.app-loading-status > small').innerHTML = messages[currentMessageIndex];
+					currentMessageIndex++;
+				}
+			}, 100);
+		}
 
 		defURI = window.location.hostname ? window.location.hostname + ':' + wsConn.port : "localhost" + ':' + wsConn.port;
 		uri = wsConn.url && /^(wss?:\/\/)?[^\/]+:[0-9]{1,5}/.test(wsConn.url) ? wsConn.url : this.buildWSURI(wsConn);
@@ -116,28 +307,48 @@ window.GoAccess = window.GoAccess || {
 		str = uri || defURI;
 		str = !/^wss?:\/\//i.test(str) ? (window.location.protocol === "https:" ? 'wss://' : 'ws://') + str : str;
 
+		if (jwt) {
+			const separator = str.includes('?') ? '&' : '?';
+			str = str + separator + 'token=' + encodeURIComponent(jwt);
+		}
+		// Store socket for token refresh
 		var socket = new WebSocket(str);
+		this.socket = socket;
+
 		socket.onopen = function (event) {
+			clearInterval(messageInterval);
+			if (this.currentJWT)
+				$('.app-loading-status > small').innerHTML = 'Authentication successful.';
+
 			this.currDelay = this.wsDelay;
 			this.retries = 0;
 
-			// attempt to keep connection alive (e.g., ping/pong)
-			if (wsConn.ping_interval)
+			if (wsConn.ping_interval) {
 				pingId = setInterval(() => { socket.send('ping'); }, wsConn.ping_interval * 1E3);
-
+			}
 			GoAccess.Nav.WSOpen(str);
+
 		}.bind(this);
 
 		socket.onmessage = function (event) {
 			this.AppState['updated'] = true;
 			this.AppData = JSON.parse(event.data);
+			if (!this.isAppInitialized) {
+				GoAccess.App.initialize();
+				GoAccess.Nav.WSOpen(str);
+				this.isAppInitialized = true;
+			}
 			this.App.renderData();
 		}.bind(this);
 
 		socket.onclose = function (event) {
+			clearInterval(messageInterval);
+			$('.app-loading-status > small').innerHTML = 'Unable to authenticate WebSocket.';
+			$('.loading-container > .spinner').style.display = 'none';
+
 			GoAccess.Nav.WSClose();
 			window.clearInterval(pingId);
-			socket = null;
+			this.socket = null;
 			this.wsTimer = setTimeout(() => { this.reconnect(wsConn); }, this.currDelay);
 		}.bind(this);
 	},
@@ -269,7 +480,7 @@ GoAccess.Util = {
 			}
 		}
 
-		return value == 0 ? String(val) : (val === undefined ? '—' : val);
+		return value == 0 ? String(val) : (val === undefined ? '-' : val);
 	},
 
 	isPanelHidden: function (panel) {
@@ -357,6 +568,35 @@ GoAccess.Util = {
 		GoAccess.Charts.initialize();
 		GoAccess.Tables.initialize();
 	},
+
+	reorderPanels: function(fromIndex, toIndex) {
+		var order = GoAccess.AppPrefs.panelOrder;
+
+		// Ensure we have a valid order array
+		if (!order || order.length === 0) {
+			console.error('Panel order not initialized');
+			return;
+		}
+
+		// Validate indices
+		if (fromIndex < 0 || fromIndex >= order.length ||
+			toIndex < 0 || toIndex >= order.length) {
+			console.error('Invalid drag indices', fromIndex, toIndex);
+			return;
+		}
+
+		// Perform the reorder
+		var item = order.splice(fromIndex, 1)[0];
+		order.splice(toIndex, 0, item);
+
+		// Save preferences
+		GoAccess.setPrefs();
+
+		// Re-render panels in new order
+		GoAccess.Panels.initialize();
+		GoAccess.Charts.initialize();
+		GoAccess.Tables.initialize();
+	},
 };
 
 // OVERALL STATS
@@ -365,25 +605,20 @@ GoAccess.OverallStats = {
 
 	// Render each overall stats box
 	renderBox: function (data, ui, row, x, idx) {
-		var wrap = $('.wrap-general-items');
+		var wrap = $('#overall ul');
+		var box = document.createElement('li');
 
-		// create a new bootstrap row every 6 elements
-		if (idx % 6 == 0) {
-			row = document.createElement('div');
-			row.setAttribute('class', 'row');
-			wrap.appendChild(row);
-		}
+		// we need to append the element first, otherwise outerHTML won't work
+		wrap.appendChild(box);
 
-		var box = document.createElement('div');
-		box.innerHTML = GoAccess.AppTpls.General.items.render({
+		box.outerHTML = GoAccess.AppTpls.General.items.render({
 			'id': x,
 			'className': ui.items[x].className,
 			'label': ui.items[x].label,
 			'value': GoAccess.Util.fmtValue(data[x], ui.items[x].dataType),
 		});
-		row.appendChild(box);
 
-		return row;
+		return wrap;
 	},
 
 	// Render overall stats
@@ -391,15 +626,16 @@ GoAccess.OverallStats = {
 		var idx = 0, row = null;
 
 		$('.last-updated').innerHTML = data.date_time;
-		$('.wrap-general').innerHTML = '';
+		$('#overall').innerHTML = '';
 
 		if (GoAccess.Util.isPanelHidden('general'))
 			return false;
 
-		$('.wrap-general').innerHTML = GoAccess.AppTpls.General.wrap.render(GoAccess.Util.merge(ui, {
+		$('#overall').innerHTML = GoAccess.AppTpls.General.wrap.render(GoAccess.Util.merge(ui, {
 			'from': data.start_date,
 			'to': data.end_date,
 		}));
+		$('#overall').setAttribute('aria-labelledby', 'overall-heading');
 
 		// Iterate over general data object
 		for (var x in data) {
@@ -516,6 +752,103 @@ GoAccess.Nav = {
 				item.classList.toggle('active');
 			}.bind(this);
 		}.bind(this));
+
+		$$('.drag-handle', function (item) {
+			var li = item.closest('li');
+
+			// Don't make the overall stats draggable
+			var link = li.querySelector('a');
+			if (link && link.getAttribute('href') === '#') {
+				return; // Skip overall stats item
+			}
+
+			li.setAttribute('draggable', 'true');
+
+			li.ondragstart = function(e) {
+				e.dataTransfer.effectAllowed = 'move';
+				e.dataTransfer.setData('text/html', this.innerHTML);
+				this.classList.add('dragging');
+
+				// Get the actual panel key from the link
+				var panelLink = this.querySelector('a');
+				var panelKey = panelLink ? panelLink.getAttribute('href').substring(1) : '';
+				e.dataTransfer.setData('panelKey', panelKey);
+
+				// Store the index in the ordered list (excluding overall)
+				var allItems = Array.from(this.parentNode.children);
+				var draggableItems = allItems.filter(function(item) {
+					var itemLink = item.querySelector('a');
+					return itemLink && itemLink.getAttribute('href') !== '#';
+				});
+				var fromIndex = draggableItems.indexOf(this);
+				e.dataTransfer.setData('index', fromIndex);
+			};
+
+			li.ondragend = function(e) {
+				this.classList.remove('dragging');
+				$$('.nav-list li', function(item) {
+					item.classList.remove('drag-over');
+				});
+			};
+
+			li.ondragover = function(e) {
+				e.preventDefault();
+				e.dataTransfer.dropEffect = 'move';
+
+				// Only allow drop on other draggable items
+				var link = this.querySelector('a');
+				if (link && link.getAttribute('href') !== '#') {
+					return false;
+				}
+				return true;
+			};
+
+			li.ondragenter = function(e) {
+				e.preventDefault();
+				// Only highlight if it's a valid drop target
+				var link = this.querySelector('a');
+				if (link && link.getAttribute('href') !== '#') {
+					this.classList.add('drag-over');
+				}
+			};
+
+			li.ondragleave = function(e) {
+				// Check if we're actually leaving the element (not just entering a child)
+				if (e.target === this) {
+					this.classList.remove('drag-over');
+				}
+			};
+
+			li.ondrop = function(e) {
+				e.stopPropagation();
+				e.preventDefault();
+
+				// Don't allow dropping on overall stats
+				var targetLink = this.querySelector('a');
+				if (targetLink && targetLink.getAttribute('href') === '#') {
+					return false;
+				}
+
+				var fromIndex = parseInt(e.dataTransfer.getData('index'));
+
+				// Calculate toIndex from draggable items only
+				var allItems = Array.from(this.parentNode.children);
+				var draggableItems = allItems.filter(function(item) {
+					var itemLink = item.querySelector('a');
+					return itemLink && itemLink.getAttribute('href') !== '#';
+				});
+				var toIndex = draggableItems.indexOf(this);
+
+				if (fromIndex !== toIndex && fromIndex !== -1 && toIndex !== -1) {
+					GoAccess.Util.reorderPanels(fromIndex, toIndex);
+					// Re-render the menu to show new order
+					GoAccess.Nav.renderMenu();
+				}
+
+				this.classList.remove('drag-over');
+				return false;
+			};
+		}.bind(this));
 	},
 
 	downloadJSON: function (e) {
@@ -579,17 +912,22 @@ GoAccess.Nav = {
 		$('html').className = '';
 		switch(theme) {
 		case 'darkGray':
+			document.querySelector('meta[name="theme-color"]')?.setAttribute('content', '#212121');
 			$('html').classList.add('dark');
 			$('html').classList.add('gray');
 			break;
 		case 'darkBlue':
+			document.querySelector('meta[name="theme-color"]')?.setAttribute('content', '#252B30');
 			$('html').classList.add('dark');
 			$('html').classList.add('blue');
 			break;
 		case 'darkPurple':
+			document.querySelector('meta[name="theme-color"]')?.setAttribute('content', '#1e1e2f');
 			$('html').classList.add('dark');
 			$('html').classList.add('purple');
 			break;
+		default:
+			document.querySelector('meta[name="theme-color"]')?.setAttribute('content', '#f0f0f0');
 		}
 		GoAccess.AppPrefs['theme'] = theme;
 		GoAccess.setPrefs();
@@ -620,12 +958,13 @@ GoAccess.Nav = {
 	},
 
 	getItems: function () {
-		var ui = GoAccess.getPanelUI(), menu = [];
+		var ui = GoAccess.getPanelUI(), menu = [], panels = [];
+
+		// Collect all valid panels
 		for (var panel in ui) {
 			if (GoAccess.Util.isPanelValid(panel))
 				continue;
-			// Push valid panels to our navigation array
-			menu.push({
+			panels.push({
 				'current': window.location.hash.substr(1) == panel,
 				'head': ui[panel].head,
 				'key': panel,
@@ -633,7 +972,32 @@ GoAccess.Nav = {
 				'hidden': GoAccess.Util.isPanelHidden(panel)
 			});
 		}
-		return menu;
+
+		// Initialize panel order if empty
+		if (!GoAccess.AppPrefs.panelOrder || GoAccess.AppPrefs.panelOrder.length === 0) {
+			GoAccess.AppPrefs.panelOrder = panels.map(function(p) { return p.key; });
+			GoAccess.setPrefs();
+		}
+
+		// Sort panels according to saved order
+		var orderedPanels = [];
+		var order = GoAccess.AppPrefs.panelOrder;
+
+		// First add panels in the saved order
+		for (var i = 0; i < order.length; i++) {
+			var panel = panels.find(function(p) { return p.key === order[i]; });
+			if (panel) orderedPanels.push(panel);
+		}
+
+		// Then add any new panels that aren't in the saved order
+		for (var j = 0; j < panels.length; j++) {
+			if (!order.includes(panels[j].key)) {
+				orderedPanels.push(panels[j]);
+				GoAccess.AppPrefs.panelOrder.push(panels[j].key);
+			}
+		}
+
+		return orderedPanels;
 	},
 
 	setPerPage: function (e) {
@@ -694,15 +1058,20 @@ GoAccess.Nav = {
 
 	WSClose: function () {
 		$$('.nav-ws-status', function (item) {
-			item.classList.remove('connected');
-			item.setAttribute('title', 'Disconnected');
+			item.classList.remove('fa-circle');
+			item.classList.add('fa-stop');
+			item.setAttribute('aria-label', GoAccess.i18n.websocket_disconnected);
+			item.setAttribute('title', GoAccess.i18n.websocket_disconnected);
 		});
 	},
 
 	WSOpen: function (str) {
+		const baseUrl = str.split('?')[0].split('#')[0];
 		$$('.nav-ws-status', function (item) {
-			item.classList.add('connected');
-			item.setAttribute('title', 'Connected to ' + str);
+			item.classList.remove('fa-stop');
+			item.classList.add('fa-circle');
+			item.setAttribute('aria-label', `${GoAccess.i18n.websocket_connected} (${baseUrl})`);
+			item.setAttribute('title', `${GoAccess.i18n.websocket_connected} (${baseUrl})`);
 		});
 	},
 
@@ -759,12 +1128,14 @@ GoAccess.Panels = {
 
 	openOpts: function (targ) {
 		var panel = targ.getAttribute('data-panel');
+    targ.setAttribute('aria-expanded', 'true');
 		targ.parentElement.classList.toggle('open');
 		this.renderOpts(panel);
 	},
 
 	closeOpts: function (e) {
 		e.currentTarget.parentElement.classList.remove('open');
+    e.currentTarget.parentElement.querySelector('[aria-expanded]').setAttribute('aria-expanded', 'false');
 		// Trigger the click event on the target if not opening another menu
 		if (e.relatedTarget && e.relatedTarget.getAttribute('data-toggle') !== 'dropdown')
 			e.relatedTarget.click();
@@ -896,6 +1267,10 @@ GoAccess.Panels = {
 		box.innerHTML = GoAccess.AppTpls.Panels.wrap.render(GoAccess.Util.merge(ui, {
 			'labels': GoAccess.i18n
 		}));
+
+		// add accessible label to parent article
+		col.setAttribute('aria-labelledby', panel);
+
 		col.appendChild(box);
 
 		// Remove pagination if not enough data for the given panel
@@ -911,15 +1286,15 @@ GoAccess.Panels = {
 		var perRow = 'horizontal' == layout ? 6 : 'wide' == layout ? 3 : 12;
 
 		// set the number of columns based on current layout
-		var col = document.createElement('div');
-		col.setAttribute('class', 'col-md-' + perRow + ' wrap-panel');
+		var col = document.createElement('article');
+		col.setAttribute('class', 'col-md-' + perRow);
 		row.appendChild(col);
 
 		return col;
 	},
 
 	createRow: function (row, idx) {
-		var wrap = $('.wrap-panels');
+		var wrap = $('#panels');
 		var layout = GoAccess.AppPrefs['layout'];
 		var every = 'horizontal' == layout ? 2 : 'wide' == layout ? 4 : 1;
 
@@ -952,16 +1327,46 @@ GoAccess.Panels = {
 	// structure.
 	renderPanels: function () {
 		var ui = GoAccess.getPanelUI(), idx = 0, row = null, col = null;
+		var order = GoAccess.AppPrefs.panelOrder || [];
 
-		$('.wrap-panels').innerHTML = '';
+		$('#panels').innerHTML = '';
+
+		// If no order is set, create default order
+		if (order.length === 0) {
+			for (var panel in ui) {
+				if (!GoAccess.Util.isPanelValid(panel) && !GoAccess.Util.isPanelHidden(panel)) {
+					order.push(panel);
+				}
+			}
+			GoAccess.AppPrefs.panelOrder = order;
+			GoAccess.setPrefs();
+		}
+
+		// Render panels in the specified order
+		for (var i = 0; i < order.length; i++) {
+			var panel = order[i];
+			if (GoAccess.Util.isPanelValid(panel) || GoAccess.Util.isPanelHidden(panel))
+				continue;
+
+			row = this.createRow(row, idx++);
+			col = this.createCol(row);
+			this.renderPanel(panel, ui[panel], col);
+		}
+
+		// Render any new panels not in the order array
 		for (var panel in ui) {
 			if (GoAccess.Util.isPanelValid(panel) || GoAccess.Util.isPanelHidden(panel))
 				continue;
-			row = this.createRow(row, idx++);
-			col = this.createCol(row);
-			// Render panel given a user interface definition
-			this.renderPanel(panel, ui[panel], col);
+			if (!order.includes(panel)) {
+				row = this.createRow(row, idx++);
+				col = this.createCol(row);
+				this.renderPanel(panel, ui[panel], col);
+				order.push(panel);
+			}
 		}
+
+		GoAccess.AppPrefs.panelOrder = order;
+		GoAccess.setPrefs();
 	},
 
 	initialize: function () {
@@ -990,8 +1395,11 @@ GoAccess.Charts = {
 		// Grab ui plot data for the selected panel
 		var plot = GoAccess.Util.getProp(GoAccess.AppState, panel + '.plot');
 
-		// Grab the data for the selected panel
-		data = data || this.processChartData(GoAccess.getPanelData(panel).data);
+		// Grab the data for the selected panel, respecting expanded state
+		if (!data) {
+			var subItems = GoAccess.Tables.getSubItemsData(panel);
+			data = this.processChartData(subItems.length ? subItems : GoAccess.getPanelData(panel).data);
+		}
 		return plot.chartReverse ? data.reverse() : data;
 	},
 
@@ -1068,7 +1476,7 @@ GoAccess.Charts = {
 
 	// Extract an array of objects that D3 can consume to process the chart.
 	// e.g., o = Object {hits: 37402, visitors: 6949, bytes:
-	// 505881789, avgts: 118609, cumts: 4436224010…}
+	// 505881789, avgts: 118609, cumts: 4436224010...}
 	processChartData: function (data) {
 		var out = [];
 		for (var i = 0; i < data.length; ++i)
@@ -1094,12 +1502,14 @@ GoAccess.Charts = {
 		return arr.join(' ');
 	},
 
-	getWMap: function (panel, plotUI, data) {
+	getWMap: function (panel, plotUI, data, projectionType) {
 		var chart = WorldMap(d3.select("#chart-" + panel));
 		chart.width($("#chart-" + panel).getBoundingClientRect().width);
 		chart.height(400);
 		chart.metric(plotUI['d3']['y0']['key']);
 		chart.opts(plotUI);
+		chart.projectionType(projectionType == 'wmap' ? 'mercator' : 'orthographic');
+		chart.panel(panel);
 
 		return chart;
 	},
@@ -1197,7 +1607,8 @@ GoAccess.Charts = {
 			chart = this.getVBar(panel, plotUI, data);
 			break;
 		case 'wmap':
-			chart = this.getWMap(panel, plotUI, data);
+		case 'gmap':
+			chart = this.getWMap(panel, plotUI, data, this.getChartType(panel));
 			break;
 		}
 
@@ -1258,11 +1669,8 @@ GoAccess.Charts = {
 
 	// Reload (doesn't redraw) the given chart's data
 	reloadChart: function (chart, panel) {
-		var subItems = GoAccess.Tables.getSubItemsData(panel);
-		var data = (subItems.length ? subItems : GoAccess.getPanelData(panel).data).slice(0);
-
 		d3.select("#chart-" + panel)
-			.datum(this.processChartData(this.getPanelData(panel, data)))
+			.datum(this.getPanelData(panel))
 			.call(chart.width($("#chart-" + panel).offsetWidth))
 			.append("div").attr("class", "chart-tooltip-wrap");
 	},
@@ -1392,71 +1800,123 @@ GoAccess.Tables = {
 	},
 
 	getSubItemsData: function (panel) {
-		var out = [], items = this.chartData[panel];
-		for (var x in items) {
-			if (!items.hasOwnProperty(x))
-				continue;
-			out = out.concat(items[x]);
+		const expanded = GoAccess.Util.getProp(GoAccess.AppState, panel + '.expanded') || {};
+		const fullData = (GoAccess.getPanelData(panel) || {}).data || [];
+
+		let results = [];
+
+		fullData.forEach(function (continent) {
+			const continentKey = GoAccess.Util.hashCode(continent.data);
+			if (!expanded[continentKey]) return;
+			if (!continent.items) return;
+
+			continent.items.forEach(function (country) {
+				const countryKey = GoAccess.Util.hashCode(country.data);
+
+				if (expanded[countryKey] && country.items) {
+					// expanded country -> show e.g., its cities
+					results = results.concat(country.items);
+				} else {
+					// not expanded -> show e.g., country
+					results.push(country);
+				}
+			});
+		});
+
+		return results;
+	},
+
+	addChartData: function (panel, keyPath) {
+		if (!keyPath) return GoAccess.getPanelData(panel).data;
+
+		const parts = keyPath.split('|');
+		let current = GoAccess.getPanelData(panel).data;
+
+		// Traverse the tree to the target node
+		for (const part of parts) {
+			const found = current.find(item => GoAccess.Util.hashCode(item.data) === part);
+			if (!found || !found.items) return [];
+			current = found.items;
 		}
-		return out;
+
+		// Store the current level's items under this exact path
+		GoAccess.Util.setProp(this.chartData, panel + '.' + keyPath, current);
+
+		// For the chart: we return **only the items at the current drill level**
+		// (not all sub-items flattened - that was the old behavior)
+		return current;
 	},
 
-	addChartData: function (panel, key) {
-		var data = this.getDataByKey(panel, key);
-		var path = panel + '.' + key;
+	removeChartData: function (panel, keyPath) {
+		// Remove this specific path
+		const path = panel + '.' + keyPath;
+		if (GoAccess.Util.getProp(this.chartData, path)) {
+			// We don't have deleteProp -> set to null / empty
+			GoAccess.Util.setProp(this.chartData, path, null);
+		}
 
-		if (!data || !data.items)
-			return [];
-		GoAccess.Util.setProp(this.chartData, path, data.items);
+		// Find the deepest remaining expanded path (or fall back to top-level)
+		const expanded = GoAccess.Util.getProp(GoAccess.AppState, panel + '.expanded') || {};
+		let deepestPath = '';
+		let maxDepth = -1;
 
-		return this.getSubItemsData(panel);
-	},
+		for (const k in expanded) {
+			if (expanded[k] !== true) continue;
+			const depth = k.split('|').length;
+			if (depth > maxDepth) {
+				maxDepth = depth;
+				deepestPath = k;
+			}
+		}
 
-	removeChartData: function (panel, key) {
-		if (GoAccess.Util.getProp(this.chartData, panel + '.' + key))
-			delete this.chartData[panel][key];
-
-		if (!this.chartData[panel] || Object.keys(this.chartData[panel]).length == 0)
-			return GoAccess.getPanelData(panel).data;
-
-		return this.getSubItemsData(panel);
+		if (deepestPath) {
+			return GoAccess.Util.getProp(this.chartData, panel + '.' + deepestPath) || GoAccess.getPanelData(panel).data;
+		}
+		return GoAccess.getPanelData(panel).data;
 	},
 
 	isExpanded: function (panel, key) {
-		var path = panel + '.expanded.' + key;
-		return GoAccess.Util.getProp(GoAccess.AppState, path);
+		// getProp returns undefined if path doesn't exist -> treat as not expanded
+		return !!GoAccess.Util.getProp(GoAccess.AppState, panel + '.expanded.' + key);
 	},
 
 	toggleExpanded: function (panel, key) {
-		var path = panel + '.expanded.' + key, ret = true;
+		var path = panel + '.expanded.' + key;
+		var currentlyExpanded = this.isExpanded(panel, key);
 
-		if (this.isExpanded(panel, key)) {
-			delete GoAccess.AppState[panel]['expanded'][key];
+		if (currentlyExpanded) {
+			// Instead of delete -> set to false or null
+			GoAccess.Util.setProp(GoAccess.AppState, path, false);
+			// or: GoAccess.Util.setProp(GoAccess.AppState, path, null);
+			// or even remove the property completely if you prefer (see Option 2)
 		} else {
-			GoAccess.Util.setProp(GoAccess.AppState, path, true), ret = false;
+			GoAccess.Util.setProp(GoAccess.AppState, path, true);
 		}
 
-		return ret;
+		return currentlyExpanded; // returns true if it WAS expanded (now collapsed)
 	},
 
 	// Toggle children rows
 	toggleRow: function (ele) {
-		var hide = false, data = [];
-		var row = ele.parentNode;
-		var panel = row.getAttribute('data-panel'), key = row.getAttribute('data-key');
-		var plotUI = GoAccess.AppCharts[panel].opts();
+		const row = ele.closest('tr');
+		if (!row) return;
+
+		const panel = row.getAttribute('data-panel');
+		let key = row.getAttribute('data-node-key') || row.getAttribute('data-key');
+		if (!key) return;
+
+		// If this is a nested row, we need to build the full path
+		// For simplicity, we'll assume nodeKey is unique enough for now.
+		// If collisions happen, we can later implement full path building.
+
+		const wasExpanded = this.toggleExpanded(panel, key);
 
-		hide = this.toggleExpanded(panel, key);
 		this.renderTable(panel, this.getCurPage(panel));
-		if (!plotUI.redrawOnExpand)
-			return;
 
-		if (!hide) {
-			data = GoAccess.Charts.processChartData(this.addChartData(panel, key));
-		} else {
-			data = GoAccess.Charts.processChartData(this.removeChartData(panel, key));
-		}
-		GoAccess.Charts.drawPlot(panel, plotUI, data);
+		const plotUI = GoAccess.AppCharts[panel]?.opts?.();
+		if (!plotUI || !plotUI.redrawOnExpand) return;
+
+		GoAccess.Charts.reloadChart(GoAccess.AppCharts[panel], panel);
 	},
 
 	// Get current panel page
@@ -1599,26 +2059,31 @@ GoAccess.Tables = {
 
 	// Given a data item object, set all the row cells and return a
 	// table row that the template can consume.
-	renderRow: function (panel, callback, ui, dataItem, idx, subItem, parentId, expanded) {
-		var shadeParent = ((!subItem && idx % 2 != 0) ? 'shaded' : '');
-		var shadeChild = ((parentId % 2 != 0) ? 'shaded' : '');
+	renderRow: function (panel, callback, ui, dataItem, idx, subItem, parentId, expanded, level) {
+		var shade = ((!subItem && idx % 2 !== 0) ? 'shaded' : '') ||
+					(subItem && (parentId % 2 !== 0) ? 'shaded' : '');
+
+		var hasChildren = !!(dataItem.items && dataItem.items.length > 0);
+
 		return {
 			'panel'       : panel,
-			'idx'         : !subItem && (String((idx + 1) + this.pageOffSet(panel))),
+			'idx'         : !subItem ? String((idx + 1) + this.pageOffSet(panel)) : '',
 			'key'         : !subItem ? GoAccess.Util.hashCode(dataItem.data) : '',
-			'expanded'    : !subItem && expanded,
+			'nodeKey'     : GoAccess.Util.hashCode(dataItem.data || dataItem),
+			'level'       : level || 0,
+			'expanded'    : !!expanded,               // must be boolean
 			'parentId'    : subItem ? String(parentId) : '',
-			'className'   : subItem ? 'child ' + shadeChild : 'parent ' + shadeParent,
-			'hasSubItems' : ui.hasSubItems,
-			'items'       : dataItem.items ? dataItem.items.length : 0,
+			'className'   : (subItem ? 'child' : 'parent') + shade,
+			'hasSubItems' : hasChildren,              // true if has children
+			'items'       : hasChildren ? dataItem.items.length : 0,
 			'cells'       : callback.call(this),
 		};
 	},
 
-	renderRows: function (rows, panel, ui, dataItems, subItem, parentId) {
+	renderRows: function(rows, panel, ui, dataItems, subItem, parentId, level = 0) {
 		subItem = subItem || false;
-		// no data rows
-		if (dataItems.length == 0 && ui.items.length) {
+		level = level || 0; /* no data rows */
+		if (dataItems.length === 0 && ui.items.length) {
 			rows.push({
 				cells: [{
 					className: 'text-center',
@@ -1626,32 +2091,34 @@ GoAccess.Tables = {
 					value: 'No data on this panel.'
 				}]
 			});
+			return;
 		}
-
-		// Iterate over all data items for the given panel and
-		// generate a table row per date item.
-		var cellcb = null;
 		for (var i = 0; i < dataItems.length; ++i) {
-			var dataItem = dataItems[i], data = null, expanded = false;
-			switch(typeof dataItem) {
-			case 'string':
-				data = dataItem;
-				cellcb = function () {
+			var dataItem = dataItems[i];
+			var data = dataItem.data || dataItem;
+			var isString = typeof dataItem === 'string';
+			var cellcb;
+			if (isString) {
+				cellcb = function() {
 					return {
 						'colspan': ui.items.length,
 						'value': data
 					};
 				};
-				break;
-			default:
-				data = dataItem.data;
+			} else {
 				cellcb = this.iterUIItems.bind(this, panel, ui.items, dataItem, this.getObjectCell.bind(this));
 			}
-
-			expanded = this.isExpanded(panel, GoAccess.Util.hashCode(data));
-			rows.push(this.renderRow(panel, cellcb, ui, dataItem, i, subItem, parentId, expanded));
-			if (dataItem.items && dataItem.items.length && expanded) {
-				this.renderRows(rows, panel, ui, dataItem.items, true, i, expanded);
+			/* Unique key for this node (important for nested expansion state) */
+			var nodeKey = !isString ? GoAccess.Util.hashCode(data) : null;
+			var expanded = nodeKey && this.isExpanded(panel, nodeKey); /* Build row with indentation level */
+			var row = this.renderRow(panel, cellcb, ui, dataItem, i, subItem, parentId, expanded); /* Add level for CSS indentation */
+			row.level = level;
+			row.nodeKey = nodeKey; /* for future use in events */
+			row.isLeaf = !(dataItem.items && dataItem.items.length > 0);
+			row.showPlaceholder = ui.hasSubItems && !row.hasSubItems;
+			rows.push(row); /* Recurse into children if expanded */
+			if (!isString && dataItem.items && dataItem.items.length && expanded) {
+				this.renderRows(rows, panel, ui, dataItem.items, true, /* is sub-item */ i, /* parent index (or could use nodeKey) */ level + 1);
 			}
 		}
 	},
@@ -1885,6 +2352,7 @@ GoAccess.App = {
 		$('nav').classList.remove('hide');
 		$('.container').classList.remove('hide');
 		$('.spinner').classList.add('hide');
+		$('.app-loading-status > small').style.display = 'none';
 
 		if (GoAccess.AppPrefs['layout'] == 'horizontal' || GoAccess.AppPrefs['layout'] == 'wide') {
 			$('.container').classList.add('container-fluid');
@@ -1915,16 +2383,20 @@ GoAccess.App = {
 		GoAccess.Tables.reloadTables();
 	},
 
-	initialize: function () {
-		this.setInitSort();
-		this.setTpls();
+	renderPanels: function () {
 		GoAccess.Nav.initialize();
-		this.initDom();
 		GoAccess.OverallStats.initialize();
 		GoAccess.Panels.initialize();
 		GoAccess.Charts.initialize();
 		GoAccess.Tables.initialize();
 	},
+
+	initialize: function () {
+		this.setInitSort();
+		this.setTpls();
+		this.initDom();
+		this.renderPanels();
+	},
 };
 
 // Adds the visibilitychange EventListener
@@ -1934,7 +2406,7 @@ document.addEventListener('visibilitychange', function () {
 		GoAccess.App.hasFocus = false;
 
 	// fires when app transitions from hidden or user returns to the app/tab.
-	if (document.visibilityState === 'visible') {
+	if (document.visibilityState === 'visible' && GoAccess.isAppInitialized) {
 		var hasFocus = GoAccess.App.hasFocus;
 		GoAccess.App.hasFocus = true;
 		hasFocus || GoAccess.App.renderData();
@@ -1950,6 +2422,5 @@ window.onload = function () {
 		'wsConnection': window.connection || null,
 		'prefs': window.html_prefs || {},
 	});
-	GoAccess.App.initialize();
 };
 }());
